#!/usr/bin/env python

import skimage.restoration
from skimage import img_as_float
import numpy as np
from PIL import Image
import os
from tqdm import tqdm
import csv
import rawpy
from utils import Color

imagesFolderPath = 'rafael/arw'
imagesFolderPathFileName = imagesFolderPath.replace('/', '_')
denoiser = 'bilateral'

raiseNotFlatFields = False
singleColorChannel = Color.GREEN_RIGHT

colors = Color if singleColorChannel is None else [singleColorChannel]

denoise = getattr(skimage.restoration, f'denoise_{denoiser}')

imagesFileNames = os.listdir(imagesFolderPath + ('/png' if raiseNotFlatFields else ''))

if raiseNotFlatFields:
    files = {}

    with open('RAISE_all.csv') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in tqdm(list(reader), 'CSV parsing'):
            file = row['File'] + '.png'
            files[file] = row

    imagesFileNames = [imageFileName for imageFileName in tqdm(imagesFileNames, 'Filtering images') if files[imageFileName]['Device'] == 'Nikon D7000' and Image.open(f'{imagesFolderPath}/png/{imageFileName}').size == (4946, 3278)]

SKY = range(2_699, 2_807)
WALL = range(2_807, 2_912)
imagesFileNames = [f'DSC0{imageIndex}.ARW' for imageIndex in SKY]

minColor = None
maxColor = None

# `color` is the actual color to estimate PRNU with.
def treatImage(imageFileName, computeExtremes = False, color = None):
    global mean, numberOfImagesInMean, minColor, maxColor
    if raiseNotFlatFields:
        imageFileName = imageFileName.replace('.png', '.NEF')
        imageFilePath = f'{imagesFolderPath}/nef/{imageFileName}'
    else:
        imageFilePath = f'{imagesFolderPath}/{imageFileName}'
    if imageFileName.endswith('.NEF') or imageFileName.endswith('.ARW'):
        with rawpy.imread(imageFilePath) as raw:
            colorDesc = raw.color_desc.decode('ascii')
            assert colorDesc == 'RGBG'
            assert np.array_equal(raw.raw_pattern, np.array([[0, 1], [3, 2]], dtype = np.uint8))
            # RG
            # GB
            rawImageVisible = raw.raw_image_visible.copy()

            redRawImageVisible = rawImageVisible[::2, ::2]
            greenRightRawImageVisible = rawImageVisible[::2, 1::2]

            greenBottomRawImageVisible = rawImageVisible[1::2, ::2]
            blueRawImageVisible = rawImageVisible[1::2, 1::2]

            match color:
                case Color.RED:
                    colorRawImageVisible = redRawImageVisible
                case Color.GREEN_RIGHT:
                    colorRawImageVisible = greenRightRawImageVisible
                case Color.GREEN_BOTTOM:
                    colorRawImageVisible = greenBottomRawImageVisible
                case Color.BLUE:
                    colorRawImageVisible = blueRawImageVisible

            if computeExtremes:
                colorRawImageVisibleMin = colorRawImageVisible.min()
                colorRawImageVisibleMax = colorRawImageVisible.max()
                if minColor is None or colorRawImageVisibleMin < minColor:
                    minColor = colorRawImageVisibleMin
                if maxColor is None or colorRawImageVisibleMax > maxColor:
                    maxColor = colorRawImageVisibleMax
                return
            imageNpArray = (colorRawImageVisible - minColor) / (maxColor - minColor)
            # Pay attention to range of values expected by the denoiser.
            # Indeed if provide the thousands valued raw image, then the denoiser only returns values between 0 and 1 and making the difference between both look pointless.
    else:
        imagePil = Image.open(imageFilePath)
        imageNpArray = img_as_float(np.array(imagePil))
    # Could make not `.nef` compatible.
    match denoiser:
        case 'wavelet':
            imageDenoisedNpArray = denoise(imageNpArray, rescale_sigma=True)
        case 'bilateral':
            imageDenoisedNpArray = denoise(imageNpArray, sigma_color=0.05, sigma_spatial=15)
        case 'tv_chambolle':
            imageDenoisedNpArray = denoise(imageNpArray, weight=0.2)
    imageNoiseNpArray = imageNpArray - imageDenoisedNpArray
    if mean is None:
        mean = imageNoiseNpArray
    else:
        mean = ((mean * numberOfImagesInMean) + imageNoiseNpArray) / (numberOfImagesInMean + 1)
    numberOfImagesInMean += 1

# Assuming same intensity scale across color channels.
for imageFileName in tqdm(imagesFileNames, 'Computing extremes of images'):
    for color in colors:
        treatImage(imageFileName, computeExtremes = True, color = color)

# To skip this step next time.
# Maybe thanks to `rawpy.RawPy` fields, possibly stating device maximal value, can avoid doing so to some extent.
print(f'{minColor=}')
print(f'{maxColor=}')

for color in colors:
    mean = None
    numberOfImagesInMean = 0

    for imageFileName in tqdm(imagesFileNames, f'Denoising images for color {color}'):
        treatImage(imageFileName, color = color)

    npArrayFilePath = f'mean_{imagesFolderPathFileName}_{denoiser}_{color}.npy'
    with open(npArrayFilePath, 'wb') as f:
        np.save(f, mean)
